/*
 * Copyright (c) 1996, 2003 VIA Networking Technologies, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *    This product includes software developed by VIA Networking Technologies, Inc.
 * 4. Neither the name of the Company nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY VIA NETWORKING TECHNOLOGIES, INC. AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * VIA NETWORKING TECHNOLOGIES, INC. OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *
 * File: if_getb.c
 *
 * Purpose: FreeBSD driver main source file.
 *
 * Author: Ryan Fu
 *
 * Date: Jul 22, 2003
 *
 * Functions:
 *      getb_intr
 *      getb_media_status
 *      getb_media_change
 *      getb_start
 *      getb_ioctl
 *      getb_watchdog
 *      getb_init
 *      getb_stop
 *      getb_init_rings
 *      getb_init_rd_ring
 *      getb_init_td_ring
 *      getb_free_td_ring
 *      getb_free_rd_ring
 *      getb_free_rings
 *      getb_init_registers
 *      getb_setup_interface
 *      getb_set_multi
 *      getb_rx_srv
 *      getb_tx_srv
 *      getb_error
 *      getb_receive_frame
 *      getb_rx_csum
 *      getb_get_opt_media_mode
 *      mac_check_media_mode
 *      getb_set_media_mode
 *      getb_print_link_status
 *      mii_check_media_mode
 *      enable_flow_control_ability
 *      mii_init
 *      getb_mii_read
 *      getb_mii_write
 *      getb_soft_reset
 *      getb_init_cam_filter
 *      getb_mac_shutdown
 *      SafeDisableMiiAutoPoll
 *      EnableMiiAutoPoll
 *
 * Revision History:
 *      07-22-2003 Ryan Fu: Add new copyright header
 *
 *
 */

#include <dev/getb/if_getb.h>



/*---------------------------- Definitions -----------------------------*/

static char getb_version[]=GETB_VERSION;

PGETB_INFO  pgetb_infos=NULL;

static CHIP_INFO chip_info_table[]= {
    { CHIP_TYPE_VT6110, GETB_ADAPTER_NAM,
        256, 1, 0x00FFFFFFUL},
    {0,NULL}
};

static GETB_VENDOR_INFO getb_devs[] = {
    { VIA_VENDOR_ID, DEVICE_ID_VT6110, PCI_ANY_ID, PCI_ANY_ID,
        &chip_info_table[0]},
    { 0, 0, 0, 0, NULL }
};


#if __FreeBSD_version >= 400000

static int  getb_probe              __P((device_t));
static int  getb_attach             __P((device_t));
static int  getb_detach             __P((device_t));
static int  getb_shutdown           __P((device_t));

static BOOL getb_get_chip           __P((device_t dev, PCHIP_INFO* ppChip));
static BOOL getb_init_info          __P((device_t dev, PGETB_INFO* ppInfo, PCHIP_INFO pChip_info));
static int  getb_get_pci_info       __P((PGETB_INFO pInfo, device_t dev));
static void getb_free_pci_info      __P((PGETB_INFO pInfo));

#elif __FreeBSD_version < 400000

static unsigned long  getb_count = 0;

static const char    *getb_probe    __P((pcici_t, pcidi_t));
static void           getb_attach   __P((pcici_t, int));
static void           getb_shutdown __P((int, void *));

static BOOL           getb_get_chip __P((pcidi_t devid, PCHIP_INFO* ppChip));
static BOOL           getb_init_info __P((int unit, PGETB_INFO* ppInfo, PCHIP_INFO pChip_info));

#endif /*__FreeBSD_version*/

static void getb_intr               __P((void *));
static void getb_media_status       __P((struct ifnet *ifp, struct ifmediareq *ifmr));
static int  getb_media_change       __P((struct ifnet *ifp));
static void getb_start              __P((struct ifnet *));
static int  getb_ioctl              __P((struct ifnet *, u_long, caddr_t));
static void getb_watchdog           __P((struct ifnet *));
static void getb_init               __P((void *));
static void getb_stop               __P((void *));

static BOOL getb_init_rings         __P((PGETB_INFO pInfo));
static BOOL getb_init_rd_ring       __P((PGETB_INFO pInfo));
static BOOL getb_init_td_ring       __P((PGETB_INFO pInfo));
static void getb_free_td_ring       __P((PGETB_INFO pInfo));
static void getb_free_rd_ring       __P((PGETB_INFO pInfo));
static void getb_free_rings         __P((PGETB_INFO pInfo));
static void getb_init_registers     __P((PGETB_INFO pInfo, GETB_INIT_TYPE InitType));
static void getb_setup_interface    __P((PGETB_INFO pInfo));
static void getb_set_multi          __P((PGETB_INFO pInfo));
static int  getb_rx_srv             __P((PGETB_INFO pInfo));
static int  getb_tx_srv             __P((PGETB_INFO pInfo));
static void getb_error              __P((PGETB_INFO pInfo, u_int32_t status));
static BOOL getb_receive_frame      __P((PGETB_INFO pInfo, int idx));
static BOOL getb_alloc_rx_buf       __P((PGETB_INFO pInfo, int idx, struct mbuf* omp));

#if __FreeBSD_version >= 400000
static __inline void getb_rx_csum   __P((uint32_t, struct mbuf* mp));
#endif

static U32  getb_get_opt_media_mode __P((PGETB_INFO pInfo));
static U32  mac_check_media_mode    __P((PGETB_INFO pInfo));
static int  getb_set_media_mode     __P((PGETB_INFO pInfo, U32 mii_status));
static void getb_print_link_status  __P((PGETB_INFO pInfo));
static U32  mii_check_media_mode    __P((PGETB_INFO pInfo));
static void enable_flow_control_ability       __P((PGETB_INFO pInfo));
static void mii_init                __P((PGETB_INFO pInfo, U32 mii_status));
static BOOL getb_mii_read           __P((PGETB_INFO pInfo, U8 byIdx, PU16 pdata));
static BOOL getb_mii_write          __P((PGETB_INFO pInfo, U8 byIdx, U16 pdata));
static BOOL getb_soft_reset         __P((PGETB_INFO pInfo));
static void getb_init_cam_filter    __P((PGETB_INFO pInfo));
static void getb_mac_shutdown       __P((PGETB_INFO pInfo));
static void SafeDisableMiiAutoPoll  __P((PGETB_INFO pInfo));
static void EnableMiiAutoPoll       __P((PGETB_INFO pInfo));

static struct mbuf* defrag_mbuf     __P((struct mbuf* mp));

#if __FreeBSD_version >= 400000

#if __FreeBSD_version >= 502000
MODULE_DEPEND(getb, pci, 1, 1, 1);
MODULE_DEPEND(getb, ether, 1, 1, 1);
#endif

static device_method_t getb_methods[] = {
   /* Device interface */
   DEVMETHOD(device_probe, getb_probe),
   DEVMETHOD(device_attach, getb_attach),
   DEVMETHOD(device_detach, getb_detach),
   DEVMETHOD(device_shutdown, getb_shutdown),
   {0, 0}
};

static driver_t getb_driver = {
   "getb", getb_methods, sizeof(GETB_INFO)
};

static devclass_t getb_devclass;

DRIVER_MODULE(getb, pci, getb_driver, getb_devclass, 0, 0);

#elif __FreeBSD_version < 400000

static struct pci_device getb_device = {
    "getb",
    getb_probe,
    getb_attach,
    &getb_count,
    NULL
};

DATA_SET(pcidevice_set, getb_device);

#endif /*__FreeBSD_version*/




/*---------------------------- Functions ----------------------------------*/


#if __FreeBSD_version >= 400000

static BOOL getb_get_chip(device_t dev, PCHIP_INFO* ppChip)
{
    U16                 wDevId, wVenId, wSubVenId, wSubDevId;
    PGETB_VENDOR_INFO   ptr;

    wVenId=pci_get_vendor(dev);

    if (wVenId!=(U16)VIA_VENDOR_ID)
        return FALSE;

    wDevId = pci_get_device(dev);
    wSubVenId = pci_get_subvendor(dev);
    wSubDevId = pci_get_subdevice(dev);

    for (ptr=&getb_devs[0]; ptr->wVendorId!=0; ptr++) {
        if ((wVenId == ptr->wVendorId) &&
            (wDevId == ptr->wDeviceId) &&
            ((wSubVenId == ptr->wSubVenId) || (ptr->wSubVenId == PCI_ANY_ID)) &&
            ((wSubDevId == ptr->wSubDevId) || (ptr->wSubDevId == PCI_ANY_ID))) {

            if (ppChip!=NULL)
                *ppChip=ptr->pChip;

            return TRUE;
         }
    }

    return FALSE;
}

static BOOL getb_init_info(device_t dev, PGETB_INFO* ppInfo, PCHIP_INFO pChip_info)
{
    PGETB_INFO  p;

    if (pgetb_infos==NULL) {
        pgetb_infos = *ppInfo;
    }
    else {
        for (p=pgetb_infos; p->next!=NULL; p=p->next)
            do {} while (0);

        p->next = *ppInfo;
        (*ppInfo)->prev=p;
    }

    (*ppInfo)->dev=dev;
    (*ppInfo)->unit=device_get_unit(dev);
    (*ppInfo)->chip_id=pChip_info->chip_id;
    (*ppInfo)->io_size=pChip_info->io_size;
    (*ppInfo)->nTxQueues=pChip_info->nTxQueue;
    (*ppInfo)->multicast_limit=MCAM_SIZE;
    (*ppInfo)->rx_buf_sz=RX_BUF_SIZE_DEFAULT;

    /*Init Options*/
    (*ppInfo)->sOpts.nTxDescs=TX_DESC_NUM_DEF;
    (*ppInfo)->sOpts.nRxDescs=(RX_DESC_NUM_DEF+3) & ~3;
    (*ppInfo)->sOpts.spd_dpx=SPEED_DUPLEX_DEF;
    (*ppInfo)->sOpts.DMA_length=DMA_LENGTH_DEF;
    (*ppInfo)->sOpts.rx_thresh=RX_THRESH_DEF;
    (*ppInfo)->sOpts.flow_cntl=FLOW_CNTL_DEF;
    (*ppInfo)->sOpts.flags=0;

#if VAL_PKTLEN_DEF == 1
    (*ppInfo)->sOpts.flags|=GETB_FLAGS_VAL_PKT_LEN;
#endif

#if HWCSUM_OFFLOAD == 1
    (*ppInfo)->sOpts.flags|=GETB_FLAGS_TX_CSUM|GETB_FLAGS_RX_CSUM;
#endif

#if __FreeBSD_version >= 500000 && VLAN_TAGGING == 1
    (*ppInfo)->sOpts.flags|=GETB_FLAGS_VLANTAG;
#endif

    (*ppInfo)->flags=(*ppInfo)->sOpts.flags;

    callout_handle_init(&((*ppInfo)->timer_handle));

    return TRUE;
}

static int  getb_get_pci_info(PGETB_INFO pInfo, device_t dev)
{
    U16 pci_cmd;
    int res_id=GETB_MEMIO_RES_ID;

    pInfo->byRevId = pci_read_config(dev, GETB_PCI_REVID, 1);
    pInfo->SubVendorID = pci_get_subvendor(dev);
    pInfo->SubSystemID = pci_get_subdevice(dev);


#if __FreeBSD_version >= 500000
    pci_cmd = pci_read_config(dev, GETB_PCI_COMMAND, 2);
    pci_enable_busmaster(dev);
#else
    pci_cmd = pci_read_config(dev, PCIR_COMMAND, 2);

    if (!((pci_cmd & PCIM_CMD_BUSMASTEREN) && (pci_cmd & PCIM_CMD_MEMEN))) {
        GETB_PRT0(pInfo, "Memory Access and/or Bus Master bits were not set!\n");
        pci_cmd |= (PCIM_CMD_BUSMASTEREN | PCIM_CMD_MEMEN);
        pci_write_config(dev, PCIR_COMMAND, pci_cmd, 2);
    }
#endif

    pInfo->mem_res = bus_alloc_resource(dev, SYS_RES_MEMORY, &res_id, 0, ~0, 1, RF_ACTIVE);

    if (pInfo->mem_res==NULL) {
        GETB_PRT0(pInfo, "Unable to allocate bus resource: memory\n");
        goto fail;
    }

    pInfo->mem_tag=rman_get_bustag(pInfo->mem_res);
    pInfo->mem_hdl=rman_get_bushandle(pInfo->mem_res);

    res_id=0;
    pInfo->int_res = bus_alloc_resource(dev, SYS_RES_IRQ, &res_id, 0, ~0, 1, RF_SHAREABLE | RF_ACTIVE);

    if (pInfo->int_res==NULL) {
        GETB_PRT0(pInfo, "Unable to allocate bus resource: interrupt\n");
        goto fail;
    }

    if (bus_setup_intr(dev, pInfo->int_res, INTR_TYPE_NET,
                       (void (*)(void *))getb_intr, pInfo,
                       &pInfo->int_hdl_tag)) {

        GETB_PRT0(pInfo, "Error registering interrupt handler!\n");
        goto fail;
    }

    return 0;

fail:
    getb_free_pci_info(pInfo);
    return(ENXIO);

}

static void getb_free_pci_info(PGETB_INFO pInfo)
{
    if (pInfo->int_hdl_tag) {
        bus_teardown_intr(pInfo->dev, pInfo->int_res, pInfo->int_hdl_tag);
    }

    if (pInfo->int_res) {
        bus_release_resource(pInfo->dev, SYS_RES_IRQ, 0, pInfo->int_res);
    }

    if (pInfo->mem_res) {
        bus_release_resource(pInfo->dev, SYS_RES_MEMORY, GETB_MEMIO_RES_ID, pInfo->mem_res);
    }
}

static int getb_probe(device_t dev)
{
    char                buf[256];
    PCHIP_INFO          pChip;

    if (getb_get_chip(dev, &pChip)) {
        sprintf(buf, "%s, Version - %s", pChip->name, getb_version);
        device_set_desc_copy(dev, buf);
        return(0);
    }

    return ENXIO;
}

#elif __FreeBSD_version < 400000

static BOOL getb_get_chip(pcidi_t devid, PCHIP_INFO* ppChip)
{
    U16                 wDevId, wVenId;
    PGETB_VENDOR_INFO   ptr;

    wVenId=(U16)(devid & 0xFFFF);

    if (wVenId!=(U16)VIA_VENDOR_ID)
        return FALSE;

    wDevId = (U16)((devid >> 16) & 0xFFFF);

    for (ptr=&getb_devs[0]; ptr->wVendorId!=0; ptr++) {
        if ((wVenId == ptr->wVendorId) && (wDevId == ptr->wDeviceId)) {
            if (ppChip!=NULL)
                *ppChip=ptr->pChip;
            return TRUE;
         }
    }

    return FALSE;
}

static BOOL getb_init_info(int unit, PGETB_INFO* ppInfo, PCHIP_INFO pChip_info)
{
    PGETB_INFO  p;

    if (pgetb_infos==NULL) {
        pgetb_infos = *ppInfo;
    }
    else {
        for (p=pgetb_infos; p->next!=NULL; p=p->next)
            do {} while (0);

        p->next = *ppInfo;
        (*ppInfo)->prev=p;
    }

    /*(*ppInfo)->dev=dev;*/
    (*ppInfo)->unit=unit;
    (*ppInfo)->chip_id=pChip_info->chip_id;
    (*ppInfo)->io_size=pChip_info->io_size;
    (*ppInfo)->nTxQueues=pChip_info->nTxQueue;
    (*ppInfo)->multicast_limit=MCAM_SIZE;
    (*ppInfo)->rx_buf_sz=RX_BUF_SIZE_DEFAULT;

    /*Init Options*/
    (*ppInfo)->sOpts.nTxDescs=TX_DESC_NUM_DEF;
    (*ppInfo)->sOpts.nRxDescs=(RX_DESC_NUM_DEF+3) & ~3;
    (*ppInfo)->sOpts.spd_dpx=SPEED_DUPLEX_DEF;
    (*ppInfo)->sOpts.DMA_length=DMA_LENGTH_DEF;
    (*ppInfo)->sOpts.rx_thresh=RX_THRESH_DEF;
    (*ppInfo)->sOpts.flow_cntl=FLOW_CNTL_DEF;
    (*ppInfo)->sOpts.flags=0;

#if VAL_PKTLEN_DEF == 1
    (*ppInfo)->sOpts.flags|=GETB_FLAGS_VAL_PKT_LEN;
#endif

    (*ppInfo)->flags=(*ppInfo)->sOpts.flags;

    /*callout_handle_init(&((*ppInfo)->timer_handle));*/

    return TRUE;
}

static const char * getb_probe(pcici_t config_id, pcidi_t device_id)
{
    PCHIP_INFO          pChip;

    if (getb_get_chip(device_id, &pChip)) {
        printf("%s, Version - %s\n", pChip->name, getb_version);
        return(pChip->name);
    }

    return(NULL);
}

#endif /*__FreeBSD_version*/



#if __FreeBSD_version >= 400000

static int getb_attach(device_t dev)
{
    int         i, error=0;
    PGETB_INFO  pInfo=device_get_softc(dev);
    PCHIP_INFO  pChip;
    int         status;

#if __FreeBSD_version >= 500000
    mtx_init(&pInfo->get_mtx, device_get_nameunit(dev), MTX_NETWORK_LOCK, MTX_DEF | MTX_RECURSE);
#else
    pInfo->get_splcount=0;
#endif

    GET_LOCK(pInfo);

    if (pInfo==NULL) {
        printf("getb: Driver infomation structure allocation failed\n");
        error=ENOMEM;
        goto fail;
    }

    /* search for chip information using PCI Device ID, Vendor ID*/
    if (!getb_get_chip(dev, &pChip)) {
        error=ENXIO;
        goto fail;
    }

    /* init GETB_INFO device private data structure */
    getb_init_info(dev, &pInfo, pChip);

    /* setup PCI resources(mem, io, irq) */
    if ((status=getb_get_pci_info(pInfo, dev))) {
        error=status;
        goto fail;
    }

    /* allocate RD/TD rings */
    if (!getb_init_rings(pInfo)) {
        getb_free_pci_info(pInfo);
        error=ENOMEM;
        goto fail;
    }

    /* init RD ring & allocate rx buffer*/
    if (!getb_init_rd_ring(pInfo)) {
        getb_free_pci_info(pInfo);
        getb_free_rings(pInfo);
        error=ENOMEM;
        goto fail;
    }

    /* init TD ring */
    if (!getb_init_td_ring(pInfo)) {
        getb_free_rd_ring(pInfo);
        getb_free_rings(pInfo);
        getb_free_pci_info(pInfo);
        error=ENOMEM;
        goto fail;
    }

    /* Reset the adapter. */
    getb_soft_reset(pInfo);

    for (i = 0; i < 6; i++)
        pInfo->arpcom.ac_enaddr[i] = readb(pInfo, MAC_REG_PAR+i);

    /* setup ifnet interface */
    getb_setup_interface(pInfo);

    pInfo->dwPHYId=MII_GET_PHY_ID(pInfo);

    /* Print out mii status */
    mac_disable_int(pInfo);
    EnableMiiAutoPoll(pInfo);
    pInfo->mii_status=mac_check_media_mode(pInfo);

    if (!(pInfo->mii_status & GETB_LINK_FAIL))
        getb_print_link_status(pInfo);

    /* Clear Hardware MIB counter */
    mac_hw_mibs_init(pInfo);

    GET_UNLOCK(pInfo);
    return 0;

fail:
    GET_UNLOCK(pInfo);
#if __FreeBSD_version >= 500000
    mtx_destroy(&pInfo->get_mtx);
#endif
    return(error);
}


static int getb_detach(device_t dev)
{
    PGETB_INFO     pInfo=device_get_softc(dev);
    struct ifnet   *ifp = &pInfo->arpcom.ac_if;

    GET_LOCK(pInfo);

    ifp->if_flags &= ~(IFF_RUNNING | IFF_OACTIVE | IFF_UP);
    ifp->if_timer = 0;

    getb_stop(pInfo);

#if __FreeBSD_version >= 400000 && __FreeBSD_version < 410000
    if_detach(ifp);
#elif __FreeBSD_version >= 410000 && __FreeBSD_version < 500000
    ether_ifdetach(ifp, ETHER_BPF_SUPPORTED);
#elif __FreeBSD_version >= 500000
    ether_ifdetach(ifp);
#endif /*__FreeBSD_version*/

    getb_free_td_ring(pInfo);
    getb_free_rd_ring(pInfo);
    getb_free_rings(pInfo);
    getb_free_pci_info(pInfo);

    GET_UNLOCK(pInfo);

#if __FreeBSD_version >= 500000
    mtx_destroy(&pInfo->get_mtx);
#endif

    return 0;
}

#elif __FreeBSD_version < 400000

static void getb_attach(pcici_t config_id, int unit)
{
    int             i;
    vm_offset_t     pbase, vbase;
    u_int32_t       pci_cmd, id;

    PGETB_INFO  pInfo;
    PCHIP_INFO  pChip = &chip_info_table[0];


    id = pci_conf_read(config_id, GETB_PCI_VENDOR_ID);

    if (PCI_VENDORID(id) != VIA_VENDOR_ID) {
        printf("getb%d: wrong vendor ID !\n", unit);
        return;
    }

    if (PCI_CHIPID(id) != DEVICE_ID_VT6110) {
        printf("getb%d: wrong device ID !\n", unit);
        return;
    }

    pInfo = (PGETB_INFO)malloc(sizeof(GETB_INFO), M_DEVBUF, M_NOWAIT);

    if (pInfo==NULL) {
        printf("getb%d: Driver infomation structure allocation failed\n", unit);
        return;
    }

    bzero(pInfo, sizeof(GETB_INFO));

    GET_LOCK(pInfo);

    /* init GETB_INFO device private data structure */
    getb_init_info(unit, &pInfo, pChip);

    /* setup PCI resources(mem, io, irq) */

    pInfo->byRevId = pci_conf_read(config_id, GETB_PCI_REVID) & 0x000000FF;

    pci_cmd = pci_conf_read(config_id, PCI_COMMAND_STATUS_REG);

    if (!((pci_cmd & PCIM_CMD_BUSMASTEREN) && (pci_cmd & PCIM_CMD_MEMEN))) {
        GETB_PRT0(pInfo, "Memory Access and/or Bus Master bits were not set!\n");
        pci_cmd |= (PCIM_CMD_BUSMASTEREN | PCIM_CMD_MEMEN);
        pci_conf_write(config_id, PCI_COMMAND_STATUS_REG, pci_cmd);
    }

    if (!pci_map_mem(config_id, GETB_PCI_LOMEM, &vbase, &pbase)) {
        printf ("getb%d: couldn't map memory\n", unit);
        goto fail;
    }

    pInfo->mem_tag=I386_BUS_SPACE_MEM;
    pInfo->mem_hdl=vbase;

    /* Allocate interrupt */
    if (!pci_map_int(config_id, getb_intr, pInfo, &net_imask)) {
        printf("getb%d: couldn't map interrupt\n", unit);
        goto fail;
    }

    /* allocate RD/TD rings */
    if (!getb_init_rings(pInfo)) {
        printf("getb%d: couldn't allocate RD/TD rings\n", unit);
        goto fail;
    }

    /* init RD ring & allocate rx buffer*/
    if (!getb_init_rd_ring(pInfo)) {
        printf("getb%d: couldn't init RD ring & allocate rx buffer\n", unit);
        getb_free_rings(pInfo);
        goto fail;
    }

    /* init TD ring */
    if (!getb_init_td_ring(pInfo)) {
        printf("getb%d: couldn't init TD ring\n", unit);
        getb_free_rd_ring(pInfo);
        getb_free_rings(pInfo);
        goto fail;
    }

    /* Reset the adapter. */
    getb_soft_reset(pInfo);

    for (i = 0; i < 6; i++)
        pInfo->arpcom.ac_enaddr[i] = readb(pInfo, MAC_REG_PAR+i);

    /* setup ifnet interface */
    getb_setup_interface(pInfo);

    pInfo->dwPHYId=MII_GET_PHY_ID(pInfo);

    /* Print out mii status */
    mac_disable_int(pInfo);
    EnableMiiAutoPoll(pInfo);
    pInfo->mii_status=mac_check_media_mode(pInfo);

    if (!(pInfo->mii_status & GETB_LINK_FAIL))
        getb_print_link_status(pInfo);

    /* Clear Hardware MIB counter */
    mac_hw_mibs_init(pInfo);

    GET_UNLOCK(pInfo);
    return;

fail:
    GET_UNLOCK(pInfo);
    free(pInfo, M_DEVBUF);
    return;
}

#endif /*__FreeBSD_version*/


#if __FreeBSD_version >= 400000

static int  getb_shutdown(device_t dev)
{
    PGETB_INFO  pInfo=device_get_softc(dev);

    getb_stop(pInfo);

    return 0;
}

#elif __FreeBSD_version < 400000

static void getb_shutdown(int howto, void * arg)
{
    PGETB_INFO  pInfo=(PGETB_INFO)arg;

    getb_stop(pInfo);

    return;
}

#endif /*__FreeBSD_version*/


static void getb_intr (void * arg)
{
    PGETB_INFO      pInfo;
    struct ifnet*   ifp;
    u_int32_t       isr_status;

    pInfo = arg;

    GET_LOCK(pInfo);
    ifp = &pInfo->arpcom.ac_if;

    if (!(ifp->if_flags & IFF_UP)) {
        GET_UNLOCK(pInfo);
        return;
    }

    /* Disable interrupts */
    mac_disable_int(pInfo);

    for (;;) {

        isr_status=mac_read_isr(pInfo);

        /* If the card has gone away the read returns 0xffff. */
        if (isr_status == 0xFFFFFFFF)
            break;

        if (isr_status)
            mac_write_isr(pInfo, isr_status);

        if ((isr_status & pInfo->IntMask) == 0)
            break;

        /*No need to check ISR status for Rx, since we'll check RD status in Rx service routine.*/
        /*if (isr_status & (ISR_PRXI|ISR_PPRXI))*/
            getb_rx_srv(pInfo);

        if (isr_status & (ISR_PTXI|ISR_PPTXI))
            getb_tx_srv(pInfo);

        if (isr_status & ~(ISR_PRXI|ISR_PPRXI|ISR_PTXI|ISR_PPTXI)) {
            getb_error(pInfo, isr_status);
        }
    }

    /* Re-enable interrupts */
    mac_enable_int(pInfo);

    GET_UNLOCK(pInfo);

    if (ifp->if_flags & IFF_RUNNING && ifp->if_snd.ifq_head != NULL)
        getb_start(ifp);

}


static int getb_rx_srv (PGETB_INFO pInfo)
{
    int                     works=0;
    PRX_DESC                pRD;
    struct  ifnet*          ifp = &(pInfo->arpcom.ac_if);
    int                     iCurrRDIdx=pInfo->iCurrRDIdx;

    while (TRUE) {

        pRD = &(pInfo->aRDRing[iCurrRDIdx]);

        if (pRD->rd0sts & OWNED_BY_NIC)
            break;

        if (!getb_receive_frame(pInfo, iCurrRDIdx))
            ifp->if_ierrors++;

        pRD->wBufLen |= RXDESC_I;

        /* Append 4 RDs each time */
        if ((iCurrRDIdx % 4) == 3) {
            int i, iPrevRDIdx = iCurrRDIdx;
            for (i=0; i<4; i++) {
                pRD = &(pInfo->aRDRing[iPrevRDIdx]);
                pRD->rd0sts |= OWNED_BY_NIC;
                if (iPrevRDIdx <= 0)
                    iPrevRDIdx = pInfo->sOpts.nRxDescs - 1;
                else
                    --iPrevRDIdx;
            }
            writew(pInfo, 4, MAC_REG_RBRDU);
        }

        iCurrRDIdx = (iCurrRDIdx+1) % pInfo->sOpts.nRxDescs;

        if (++works == pInfo->sOpts.nRxDescs)
            break;

    } /* while */

    pInfo->iCurrRDIdx=iCurrRDIdx;

    return works;
}


#if __FreeBSD_version >= 400000
static __inline void getb_rx_csum(uint32_t rxctl, struct mbuf* mp)
{
    mp->m_pkthdr.csum_flags=0;

    /* Check IP header checksum */
    if (rxctl & CSM_IPKT)
        mp->m_pkthdr.csum_flags |= CSUM_IP_CHECKED;
    if (rxctl & CSM_IPOK)
        mp->m_pkthdr.csum_flags |= CSUM_IP_VALID;

    /* Check TCP/UDP checksum */
    if (rxctl & (CSM_TCPKT|CSM_UDPKT) &&
        rxctl & CSM_TUPOK) {
        mp->m_pkthdr.csum_flags |= CSUM_DATA_VALID|CSUM_PSEUDO_HDR;
        mp->m_pkthdr.csum_data = 0xffff;
    }
}
#endif

static BOOL getb_receive_frame(PGETB_INFO pInfo, int idx)
{
    PGETB_RD_INFO           pRDInfo = &(pInfo->aRDInfo[idx]);
    PRX_DESC                pRD = &(pInfo->aRDRing[idx]);
    struct  ifnet*          ifp = &(pInfo->arpcom.ac_if);
    struct ether_header*    eh;
    struct mbuf*            mp;
    int                     total_len;
    uint32_t                rxstat, rxctl;

    rxstat = pRD->rd0sts;
    rxctl = pRD->rd1ctl;

    /*
     * Bad/error frames will have the RXOK bit cleared.
     * However, there's one error case we want to allow:
     * if a VLAN tagged frame arrives and the chip can't
     * match it against the CAM filter, it considers this
     * a 'VLAN CAM filter miss' and clears the 'RXOK' bit.
     * We don't want to drop the frame though: our VLAN
     * filtering is done in software.
     */
    if (!(rxstat & RSR_RXOK)) {
        if (rxstat & (RSR_RL|RSR_CE|RSR_VIDM))  {
            /*GETB_PRT0(pInfo, "Don't drop CE and RL packet.\n");*/
        }
        else { /*Drop this RD, move to next RD*/
            /*GETB_PRT0(pInfo, "Drop error packet.\n");*/
            return FALSE;
        }
    }

    if (rxstat & (RSR_STP|RSR_EDP)) {
        GETB_PRT0(pInfo, "The received frame span multple RDs\n");
        /*pStats->rx_length_errors++;*/
        return FALSE;
    }



    if (rxstat & RSR_MAR)
        ifp->if_imcasts++;

    if (rxstat & RSR_BAR)
        /*pInfo->adwRMONStats[RMON_BroadcastPkts]++*/ ;

    mp = pRDInfo->mbuf;
    total_len = ((rxstat & RDSTS_BUFSIZ) >> 16) - 4;

   /*
    * If allocating a replacement mbuf fails,
    * drop the packet
    */
    if (!getb_alloc_rx_buf(pInfo, idx, NULL)) {
        GETB_PRT0(pInfo, "Can not allocate rx buf\n");
        return FALSE;
    }

    mp->m_pkthdr.rcvif = ifp;
    mp->m_pkthdr.len = mp->m_len = total_len;

    eh = mtod(mp, struct ether_header *);

    /*drop frame not met IEEE 802.3*/
    if (pInfo->flags & GETB_FLAGS_VAL_PKT_LEN) {
        if ((htons(eh->ether_type)>=46) &&
            (htons(eh->ether_type)<=1500)) {
            U16 wSAP=*(((PU16) &eh->ether_type)+1);
            if (wSAP!=0xffff) {
                if (htons(eh->ether_type)!=mp->m_len) {
                    m_freem(mp);
                    return FALSE;
                }
            }
        }
    }

#if __FreeBSD_version >= 400000
    if (pInfo->flags & GETB_FLAGS_RX_CSUM) {
        /* Do Checksum */
        getb_rx_csum(rxctl, mp);
    }
#endif

#if __FreeBSD_version >= 500000
    if (pInfo->flags & GETB_FLAGS_VLANTAG) {
        /* Do VLAN */
        if (rxstat & RSR_VTAG) {
            /* get the tagging info */
            VLAN_INPUT_TAG(ifp, mp, ntohs(rxctl & RDCTL_PQTAG), /*errcase empty*/);
        }
    }
#endif

    ifp->if_ipackets++;
    ifp->if_ibytes += mp->m_pkthdr.len;


#if __FreeBSD_version < 400000

#if NBPFILTER > 0
        /*
         * Handle BPF listeners. Let the BPF user see the packet, but
         * don't pass it up to the ether_input() layer unless it's
         * a broadcast packet, multicast packet, matches our ethernet
         * address or the interface is in promiscuous mode.
         */
        if (ifp->if_bpf) {
            bpf_mtap(ifp, mp);
            if ((ifp->if_flags & IFF_PROMISC) &&
                (bcmp(eh->ether_dhost, sc->arpcom.ac_enaddr, ETHER_ADDR_LEN) &&
                (eh->ether_dhost[0] & 1) == 0)) {
                m_freem(mp);
                return TRUE;
            }
        }
#endif /*NBPFILTER*/

#elif __FreeBSD_version >=400000 && __FreeBSD_version < 410000
        /*
         * Handle BPF listeners. Let the BPF user see the packet, but
         * don't pass it up to the ether_input() layer unless it's
         * a broadcast packet, multicast packet, matches our ethernet
         * address or the interface is in promiscuous mode.
         */
        if (ifp->if_bpf) {
            bpf_mtap(ifp, mp);
            if ((ifp->if_flags & IFF_PROMISC) &&
                (bcmp(eh->ether_dhost, sc->arpcom.ac_enaddr, ETHER_ADDR_LEN) &&
                (eh->ether_dhost[0] & 1) == 0)) {
                m_freem(mp);
                return TRUE;
            }
        }
#endif /*__FreeBSD_version*/


#if __FreeBSD_version >= 500000
    GET_UNLOCK(pInfo);
    (*ifp->if_input)(ifp, mp);
    GET_LOCK(pInfo);
#else
    /* Remove header from mbuf and pass it on. */
    m_adj(mp, sizeof(struct ether_header));
    ether_input(ifp, eh, mp);
#endif /*__FreeBSD_version >= 500000*/

    return TRUE;
}


static int getb_tx_srv (PGETB_INFO pInfo)
{
    PTX_DESC        pTD;
    int             iQNo;
    int             idx;
    PGETB_TD_INFO   pTDInfo;
    struct  ifnet*  ifp = &(pInfo->arpcom.ac_if);


    for (iQNo=0; iQNo<pInfo->nTxQueues; iQNo++) {

        idx=pInfo->aiTailTDIdx[iQNo];

        while (idx != pInfo->aiCurrTDIdx[iQNo]) {

            pTD = &(pInfo->apTDRings[iQNo][idx]);
            pTDInfo = &(pInfo->apTDInfos[iQNo][idx]);

            if (pTD->td0sts & OWNED_BY_NIC)
                break;

            if (pTD->td0sts & TSR_TERR) {
                ifp->if_oerrors++;

#if __FreeBSD_version < 400000
                if (pTD->td0sts & TSR_OWC)
                    ifp->if_collisions++;
#endif
            }
            else {
                ifp->if_opackets++;
                ifp->if_obytes += pTDInfo->mbuf->m_pkthdr.len;
            }

#if __FreeBSD_version < 400000
            if (pTD->td0sts & TSR_COLS)
                ifp->if_collisions+=(pTD->td0sts & 0x000F);
#endif

            if (pTDInfo->mbuf) {
                m_freem(pTDInfo->mbuf);
                pTDInfo->mbuf=NULL;
            }

            pTD->aTdBufs[0].wBufSize &= ~TXDESC_Q;

            pInfo->iTDUsed[iQNo]--;
            idx = (idx+1) % pInfo->sOpts.nTxDescs;

        } /* while */

        if (idx != pInfo->aiTailTDIdx[iQNo]) {
                pInfo->aiTailTDIdx[iQNo] = idx;
                ifp->if_flags &= ~IFF_OACTIVE;
                ifp->if_timer = 0;
        }

    } /* for */

    return 0;
}


static void getb_error(PGETB_INFO pInfo, u_int32_t status)
{

    if (status & ISR_TXSTLI) {
        printf("TD DMA error!\n");
        getb_init(pInfo);
    }

    if (status & ISR_SRCI) {

        pInfo->mii_status=mac_check_media_mode(pInfo);
        getb_print_link_status(pInfo);

        enable_flow_control_ability(pInfo);

        /* This is patch for VT3119A1
         * Use TCR_TB2BDIS to prevent 1c_Reset failure in half-duplex mode
         * Only use this in 10HD and 100HD
         */
        if (pInfo->byRevId < REV_ID_VT3216_A0) {
            if (pInfo->mii_status & (GETB_SPEED_100|GETB_SPEED_10)) {
                if (pInfo->mii_status & GETB_DUPLEX_FULL) {
                    BYTE_REG_BITS_OFF(pInfo, TCR_TB2BDIS, MAC_REG_TCR);
                }
                else {
                    BYTE_REG_BITS_ON(pInfo, TCR_TB2BDIS, MAC_REG_TCR);
                }
            }
        }

        /* This is patch for VT3119/VT3216
         * Patch HW MIB TXSQEErr Bug, this should only function in 10HD
         */
        if (pInfo->byRevId < REV_ID_VT3284_A0) {
            if ((pInfo->mii_status & GETB_SPEED_10) && !(pInfo->mii_status & GETB_DUPLEX_FULL)) {
                /* Enable count TXSQEErr in 10HD */
                BYTE_REG_BITS_OFF(pInfo, 0x80, MAC_REG_TESTCFG);
            }
            else {
                /* Disable count TXSQEErr in 10FD, 100HD, 100FD, 1G */
                BYTE_REG_BITS_ON(pInfo, 0x80, MAC_REG_TESTCFG);
            }
        }

    }

    if (status & ISR_MIBFI)
        getb_update_hw_mibs(pInfo);

    if (status & ISR_LSTEI) {
        getb_rx_srv(pInfo);
        mac_rx_queue_run(pInfo);
        mac_rx_queue_wake(pInfo);
    }
}


static struct mbuf* defrag_mbuf(struct mbuf* mp)
{
    struct mbuf* m_new;

#if __FreeBSD_version >= 502000
    m_new = m_defrag(mp, M_DONTWAIT);
#else
    MGETHDR(m_new, M_DONTWAIT, MT_DATA);

    if (m_new == NULL)
        return NULL;

    if (mp->m_pkthdr.len > MHLEN) {
        MCLGET(m_new, M_DONTWAIT);
        if (!(m_new->m_flags & M_EXT)) {
            m_freem(m_new);
            return NULL;
        }
    }

    m_copydata(mp, 0, mp->m_pkthdr.len, mtod(m_new, caddr_t));
    m_new->m_pkthdr.len = m_new->m_len = mp->m_pkthdr.len;
    m_freem(mp);
#endif

    return m_new;
}

static void getb_start(struct ifnet* ifp)
{
    PGETB_INFO      pInfo=ifp->if_softc;
    int             iQNo=0;
    struct mbuf*    m_head;
    struct mbuf*    mp;
    PTX_DESC        pTD, pPreTD;
    PGETB_TD_INFO   pTDInfo;
    int             iPreTDIdx, iCurrTDIdx;
    int             iSegments=0;
    struct mbuf     *m_new = NULL;

    GET_LOCK(pInfo);

    iCurrTDIdx=pInfo->aiCurrTDIdx[iQNo];

    while (ifp->if_snd.ifq_head != NULL) {

        IF_DEQUEUE(&ifp->if_snd, m_head);

        if (m_head == NULL)
            break;

        if (AVAIL_TD(pInfo, iQNo)<=1) {
            getb_tx_srv(pInfo);
        }

        if (AVAIL_TD(pInfo, iQNo)<=1)  {
            ifp->if_flags |= IFF_OACTIVE;
            IF_PREPEND(&ifp->if_snd, m_head);
            break;
        }

        pTD = &(pInfo->apTDRings[iQNo][iCurrTDIdx]);
        pTDInfo = &(pInfo->apTDInfos[iQNo][iCurrTDIdx]);

        pTD->td1ctl = TCR_TIC;
        pTD->td1ctl |= TCPLS_NORMAL;
        pTD->aTdBufs[0].wBufSize &= ~TXDESC_Q;

#if __FreeBSD_version >= 400000
        /* Do Checksum */
        if (pInfo->flags & GETB_FLAGS_TX_CSUM) {
            if (m_head->m_pkthdr.csum_flags & CSUM_IP)
                pTD->td1ctl |= TCR_IPCK;
            if (m_head->m_pkthdr.csum_flags & CSUM_TCP)
                pTD->td1ctl |= TCR_TCPCK;
            if (m_head->m_pkthdr.csum_flags & CSUM_UDP)
                pTD->td1ctl |= TCR_UDPCK;
        }
#endif

#if __FreeBSD_version >= 500000
        /* Do VLAN */
        if (pInfo->flags & GETB_FLAGS_VLANTAG) {
            struct m_tag *mtag;
            mtag = VLAN_OUTPUT_TAG(&pInfo->arpcom.ac_if, m_head);
            if (mtag != NULL) {
                pTD->td1ctl |= (u_int32_t)VLAN_TAG_VALUE(mtag);
                pTD->td1ctl |= TCR_VETAG;
            }
        }
#endif

        /*
         * Load mbufs to TDs
         */
        if (m_head->m_pkthdr.len < GETB_MIN_FRAME_SIZE) {
            /*
             * Frame size < GETO_MIN_FRAME_SIZE
             */

            m_new = defrag_mbuf(m_head);
            if (m_new == NULL) {
                IF_PREPEND(&ifp->if_snd, m_head);
                printf("no memory for tx list");
                break;
            }

            m_head = m_new;

            /*
             * we have to make sure to pad short frames out to the minimum frame length
             * ourselves.
             */
            bzero(mtod(m_new, char*)+m_new->m_pkthdr.len, GETB_MIN_FRAME_SIZE - m_new->m_len);
            m_new->m_pkthdr.len += GETB_MIN_FRAME_SIZE - m_new->m_len;
            m_new->m_len = m_new->m_pkthdr.len;

            pTD->aTdBufs[iSegments].dwBufAddrLo=vtophys(mtod(m_new, caddr_t));
            pTD->aTdBufs[iSegments].wBufAddrHi=0;
            pTD->aTdBufs[iSegments].wBufSize=m_new->m_len;
            pTD->td0sts = (uint32_t)m_new->m_len << 16;
            iSegments=1;

        }
        else {
            /*
             * Frame size >= GETO_MIN_FRAME_SIZE
             */
again:
            pTD->td0sts = (uint32_t)m_head->m_pkthdr.len << 16;

            for (mp = m_head, iSegments=0; mp != NULL; mp = mp->m_next) {

                if (iSegments>=CB_MAX_SEGMENT)
                    break;

                if (mp->m_len == 0)
                    continue;

                pTD->aTdBufs[iSegments].dwBufAddrLo=vtophys(mtod(mp, vm_offset_t));
                pTD->aTdBufs[iSegments].wBufAddrHi=0;
                pTD->aTdBufs[iSegments].wBufSize=mp->m_len;

                iSegments++;
            }

            if (mp!=NULL) {
                /*
                 * We ran out of segments. We have to recopy this
                 * mbuf chain first. Bail out if we can't get the
                 * new buffers.
                 */
                m_new = defrag_mbuf(m_head);
                if (m_new == NULL) {
                    IF_PREPEND(&ifp->if_snd, m_head);
                    printf("no memory for tx list");
                    break;
                }

                m_head = m_new;
                goto again;
            }

        } /*else of if (m_head->m_pkthdr.len < GETB_MIN_FRAME_SIZE) */


        pTDInfo->mbuf = m_head;
        pTD->td1ctl |= ((uint32_t)iSegments+1) << 28;

        /*
         *Send a copy of the frame to the BPF listener
         */
#if __FreeBSD_version >= 500000

        BPF_MTAP(ifp, m_head);

#elif __FreeBSD_version >= 400000

        if (ifp->if_bpf)
            bpf_mtap(ifp, m_head);

#else

#if NBPFILTER > 0
        if (ifp->if_bpf)
            bpf_mtap(ifp, m_head);
#endif

#endif /*__FreeBSD_version*/


        /*
         * Set Que/Own bits
         */
        iPreTDIdx = iCurrTDIdx-1;
        if (iPreTDIdx<0)
            iPreTDIdx=pInfo->sOpts.nTxDescs-1;

        pPreTD = &(pInfo->apTDRings[iQNo][iPreTDIdx]);

        pTD->td0sts |= OWNED_BY_NIC;
        pPreTD->aTdBufs[0].wBufSize |= TXDESC_Q;

        pInfo->iTDUsed[iQNo]++;
        iCurrTDIdx = (iCurrTDIdx+1) % pInfo->sOpts.nTxDescs;

    } /* end of while loop */


    if (iCurrTDIdx == pInfo->aiCurrTDIdx[iQNo]) {
        GET_UNLOCK(pInfo);
        return;
    }

    /* Set Tx queue wake */
    mac_tx_queue_wake(pInfo, iQNo);

    pInfo->aiCurrTDIdx[iQNo] = iCurrTDIdx;

    GET_UNLOCK(pInfo);

    ifp->if_timer = GETB_TX_TIMEOUT;

    return;
}


static int  getb_ioctl(struct ifnet* ifp, u_long cmd, caddr_t data)
{
    int             error = 0;
    struct ifreq*   ifr = (struct ifreq *)data;
    PGETB_INFO      pInfo = ifp->if_softc;

    GET_LOCK(pInfo);

    switch (cmd) {
    case SIOCSIFADDR:
    case SIOCGIFADDR:
        ether_ioctl(ifp, cmd, data);
        break;

    case SIOCSIFMTU:
        if (ifr->ifr_mtu > GETB_MAX_MTU) {
            error = EINVAL;
        } else {
            if (ifr->ifr_mtu + 14 + 4 <=MCLBYTES) {
                ifp->if_mtu = ifr->ifr_mtu;
                pInfo->rx_buf_sz = ifp->if_mtu + 14+4;
                getb_init(pInfo);
            }
            else
                error=EINVAL;
        }
        break;

    case SIOCSIFFLAGS:
        if (ifp->if_flags & IFF_UP) {
            if (ifp->if_flags & IFF_RUNNING && ifp->if_flags & IFF_PROMISC) {
                writel(pInfo, 0xffffffff, MAC_REG_MARCAM);
                writel(pInfo, 0xffffffff, MAC_REG_MARCAM+4);
                BYTE_REG_BITS_ON(pInfo, RCR_AM|RCR_PROM, MAC_REG_RCR);
            } else if (ifp->if_flags & IFF_RUNNING && !(ifp->if_flags & IFF_PROMISC)) {
                BYTE_REG_BITS_OFF(pInfo, RCR_PROM, MAC_REG_RCR);
            } else
                getb_init(pInfo);
        } else {
            if (ifp->if_flags & IFF_RUNNING) {
                getb_stop(pInfo);
            }
        }
        break;

    case SIOCADDMULTI:
    case SIOCDELMULTI:
        getb_set_multi(pInfo);
        break;

    case SIOCSIFMEDIA:
    case SIOCGIFMEDIA:
        error = ifmedia_ioctl(ifp, ifr, &pInfo->media, cmd);
        break;

#if __FreeBSD_version >= 400000
    case SIOCSIFCAP:
        /* Set Checksum capability by request */
        ifp->if_capenable &= ~(IFCAP_HWCSUM);
        ifp->if_capenable |= ifr->ifr_reqcap & (IFCAP_HWCSUM);

        if (ifp->if_capenable & IFCAP_TXCSUM) {
            ifp->if_hwassist = (CSUM_IP | CSUM_TCP | CSUM_UDP);
            pInfo->flags |= GETB_FLAGS_TX_CSUM;
        }
        else {
            ifp->if_hwassist = 0;
            pInfo->flags &= ~GETB_FLAGS_TX_CSUM;
        }

        if (ifp->if_capenable & IFCAP_RXCSUM) {
            pInfo->flags |= GETB_FLAGS_RX_CSUM;
        }
        else {
            pInfo->flags &= ~GETB_FLAGS_RX_CSUM;
        }

        if (ifp->if_flags & IFF_RUNNING)
            getb_init(pInfo);

        break;
#endif

    default:
        error = EINVAL;

    } /* switch */

    GET_UNLOCK(pInfo);
    return(error);
}


static void getb_watchdog(struct ifnet* ifp)
{
        PGETB_INFO pInfo;

        pInfo = ifp->if_softc;

        GET_LOCK(pInfo);

        printf("getb%d: watchdog timeout\n", pInfo->unit);
        ifp->if_oerrors++;

        getb_tx_srv(pInfo);
        getb_rx_srv(pInfo);

        getb_init(pInfo);

        GET_UNLOCK(pInfo);

        return;
}

#if __FreeBSD_version >= 400000
static void getb_timer(void* arg)
{
    PGETB_INFO  pInfo=arg;
    struct ifnet   *ifp = &pInfo->arpcom.ac_if;

    GET_LOCK(pInfo);

    getb_update_hw_mibs(pInfo);

    ifp->if_collisions += pInfo->adwHWMIBCounters[HW_MIB_ifTxEtherCollisions];

    bzero(pInfo->adwHWMIBCounters, sizeof(pInfo->adwHWMIBCounters));
    pInfo->timer_handle = timeout(getb_timer, pInfo, 2*hz);

    GET_UNLOCK(pInfo);
}
#endif

static void getb_init(void *arg)
{
    PGETB_INFO      pInfo=(PGETB_INFO)arg;
    struct ifnet    *ifp;

    GET_LOCK(pInfo);

    getb_stop(pInfo);

    if (!getb_init_rd_ring(pInfo)) {
        GETB_PRT0(pInfo, "getb_init_rd_ring() failed.");
        GET_UNLOCK(pInfo);
        return;
    }

    if (!getb_init_td_ring(pInfo)) {
        GETB_PRT0(pInfo, "getb_init_td_ring() failed.");
        getb_free_rd_ring(pInfo);
        GET_UNLOCK(pInfo);
        return;
    }

    /*
     Initialiation of MAC registers
     */
    getb_init_registers(pInfo, GETB_INIT_COLD);

    ifp = &(pInfo->arpcom.ac_if);

    ifp->if_flags |= IFF_RUNNING;
    ifp->if_flags &= ~IFF_OACTIVE;

#if __FreeBSD_version >= 400000
    pInfo->timer_handle = timeout(getb_timer, pInfo, 2*hz);
#endif

    mac_write_int_mask(pInfo, pInfo->IntMask);

    mac_enable_int(pInfo);

    GET_UNLOCK(pInfo);
}


static void getb_stop(void *arg)
{
    PGETB_INFO    pInfo = (PGETB_INFO)arg;
    struct ifnet  *ifp = &(pInfo->arpcom.ac_if);

    GET_LOCK(pInfo);

    ifp->if_flags &= ~(IFF_RUNNING | IFF_OACTIVE);
    ifp->if_timer = 0;

    getb_mac_shutdown(pInfo);

#if __FreeBSD_version >= 400000
    untimeout(getb_timer, pInfo, pInfo->timer_handle);
#endif

    getb_free_rd_ring(pInfo);
    getb_free_td_ring(pInfo);

    GET_UNLOCK(pInfo);

    return;
}


/**********************************************************************/
static void getb_mac_shutdown(PGETB_INFO pInfo)
{
    int     i;

    mac_disable_int(pInfo);
    writel(pInfo, CR0_STOP, MAC_REG_CR0_SET);
    writew(pInfo, 0xFFFF, MAC_REG_TDCSR_CLR);
    writeb(pInfo, 0xFF, MAC_REG_RDCSR_CLR);
    writel(pInfo, 0, MAC_REG_RDBASE_LO);

    for (i=0; i<pInfo->nTxQueues; i++)
        writel(pInfo, 0, MAC_REG_TDBASE_LO+(i*4));

    SafeDisableMiiAutoPoll(pInfo);
    mac_clear_isr(pInfo);
}


static BOOL getb_init_rings(PGETB_INFO pInfo)
{
    int     i;

#if __FreeBSD_version < 400000
    unsigned int        round, pool_sz, rdpool_sz, tdpool_sz;
    caddr_t             roundptr;
    int j;

    rdpool_sz = pInfo->sOpts.nRxDescs * sizeof(RX_DESC);
    tdpool_sz = pInfo->sOpts.nTxDescs * sizeof(TX_DESC);

#endif



#if __FreeBSD_version >= 400000

    /*allocate all RD/TD rings a single pool*/

    pInfo->pool_size=(pInfo->sOpts.nRxDescs * sizeof(RX_DESC) +
                      pInfo->sOpts.nTxDescs * sizeof(TX_DESC)*pInfo->nTxQueues);

    pInfo->pool=contigmalloc(pInfo->pool_size, M_DEVBUF, M_NOWAIT, 0, ~0UL, GETB_DMA_ALIGN, 0);

    if (pInfo->pool==NULL)  {
        GETB_PRT0(pInfo, "Allocate dma memory failed\n");
        return FALSE;
    }

    bzero(pInfo->pool, pInfo->pool_size);

    pInfo->aRDRing=(PRX_DESC)pInfo->pool;

    pInfo->rd_pool_dma=(void*)vtophys((vm_offset_t)pInfo->pool);

    for (i=0; i<pInfo->nTxQueues; i++) {

        pInfo->td_pool_dma[i]=(PU8)pInfo->rd_pool_dma+
                              pInfo->sOpts.nRxDescs*sizeof(RX_DESC)+
                              pInfo->sOpts.nTxDescs*sizeof(TX_DESC)*i;

        pInfo->apTDRings[i]=(PTX_DESC)((PU8)pInfo->pool+
                            pInfo->sOpts.nRxDescs*sizeof(RX_DESC)+
                            pInfo->sOpts.nTxDescs*sizeof(TX_DESC)*i);
    }

#elif __FreeBSD_version < 400000

    for (j=0; j<pInfo->nTxQueues+1; j++) {

        if (j==0) pool_sz = rdpool_sz;
        else      pool_sz = tdpool_sz;

        pInfo->rawpool[j] = malloc(pool_sz + GETB_DMA_ALIGN, M_DEVBUF, M_NOWAIT);

        if (pInfo->rawpool[j] == NULL) {
            GETB_PRT0(pInfo, "Allocate dma memory failed\n");
            return FALSE;
        }

        round = (unsigned int)pInfo->rawpool[j] & 0xFF;
        roundptr = pInfo->rawpool[j];

        for (i = 0; i < GETB_DMA_ALIGN; i++) {
            if (round % GETB_DMA_ALIGN) {
                round++;
                roundptr++;
            } else
                break;
        }

        pInfo->pool[j] = roundptr;
        bzero(pInfo->pool[j], pool_sz);
    }

    pInfo->aRDRing=(PRX_DESC)pInfo->pool[0];

    pInfo->rd_pool_dma=(void*)vtophys((vm_offset_t)pInfo->pool[0]);

    for (i=0; i<pInfo->nTxQueues; i++) {
        pInfo->apTDRings[i]=(PTX_DESC)pInfo->pool[i+1];
        pInfo->td_pool_dma[i]=(void*)vtophys((vm_offset_t)pInfo->pool[i+1]);
    }

#endif

    return TRUE;
}


static void getb_free_rings(PGETB_INFO pInfo)
{
#if __FreeBSD_version < 400000
    int i;

    for (i=0; i<pInfo->nTxQueues+1; i++) {
        if (pInfo->pool[i]) {
            free(pInfo->rawpool[i], M_DEVBUF);
            pInfo->pool[i]=NULL;
        }
    }

#else
    if (pInfo->pool) {
        contigfree(pInfo->pool, pInfo->pool_size, M_DEVBUF);
        pInfo->pool=NULL;
    }
#endif

}


static BOOL getb_init_rd_ring(PGETB_INFO pInfo)
{
    int i;
    PRX_DESC        pDesc;
    PGETB_RD_INFO   pRDInfo;

    pInfo->aRDInfo=malloc(sizeof(GETB_RD_INFO)*pInfo->sOpts.nRxDescs, M_DEVBUF, M_NOWAIT);

    if (pInfo->aRDInfo==NULL) {
        GETB_PRT0(pInfo, "Allocate RX desc info failed\n");
        return FALSE;
    }

    bzero(pInfo->aRDInfo, sizeof(GETB_RD_INFO)*pInfo->sOpts.nRxDescs);

    /* Init the RD ring entries */
    for (i = 0; i < pInfo->sOpts.nRxDescs; i++) {
        pDesc = &(pInfo->aRDRing[i]);
        pRDInfo = &(pInfo->aRDInfo[i]);

        if (!getb_alloc_rx_buf(pInfo, i, NULL)) {
            GETB_PRT0(pInfo, "Allocate rx buf failed.\n");
            return FALSE;
        }

        pDesc->rd0sts |= OWNED_BY_NIC;
    }

    pInfo->iCurrRDIdx=0;

    return TRUE;
}


static void getb_free_rd_ring(PGETB_INFO pInfo)
{
    int i;

    if (pInfo->aRDInfo==NULL)
        return;

    for (i = 0; i < pInfo->sOpts.nRxDescs; i++) {
        PGETB_RD_INFO   pRDInfo = &(pInfo->aRDInfo[i]);
        if (pRDInfo->mbuf) {
            m_freem(pRDInfo->mbuf);
            pRDInfo->mbuf=NULL;
        }
    }

    if (pInfo->aRDInfo) {
        free(pInfo->aRDInfo, M_DEVBUF);
        pInfo->aRDInfo=NULL;
    }
}


static BOOL getb_init_td_ring(PGETB_INFO pInfo)
{
    int j;

    /* Init the TD ring entries */
    for (j=0; j<pInfo->nTxQueues; j++) {
        pInfo->apTDInfos[j]=malloc(sizeof(GETB_TD_INFO)*pInfo->sOpts.nTxDescs, M_DEVBUF, M_NOWAIT);

        if (pInfo->apTDInfos[j]==NULL) {
            GETB_PRT0(pInfo, "Allocate TX DESC failed.\n");
            return FALSE;
        }

        bzero(pInfo->apTDInfos[j], sizeof(GETB_TD_INFO)*pInfo->sOpts.nTxDescs);
        pInfo->iTDUsed[j]=pInfo->aiTailTDIdx[j]=pInfo->aiCurrTDIdx[j]=0;
    }

    return TRUE;
}


static void getb_free_td_ring(PGETB_INFO pInfo)
{
    int i, j;

    for (j=0; j<pInfo->nTxQueues; j++) {

        if (pInfo->apTDInfos[j]==NULL)
            continue;

        for (i = 0; i < pInfo->sOpts.nTxDescs; i++) {
            PGETB_TD_INFO   pTDInfo = &(pInfo->apTDInfos[j][i]);

            if (pTDInfo==NULL)
                continue;

            if (pTDInfo->mbuf) {
                m_freem(pTDInfo->mbuf);
                pTDInfo->mbuf=NULL;
            }
        }

        free(pInfo->apTDInfos[j], M_DEVBUF);
        pInfo->apTDInfos[j]=NULL;
    }
}

static BOOL getb_alloc_rx_buf(PGETB_INFO pInfo, int idx, struct mbuf* omp)
{
    PRX_DESC        pRD = &(pInfo->aRDRing[idx]);
    PGETB_RD_INFO   pRDInfo = &(pInfo->aRDInfo[idx]);
    struct mbuf*    mp;

    if (omp==NULL) {
#if __FreeBSD_version >= 502000
        mp = m_getcl(M_DONTWAIT, MT_DATA, M_PKTHDR);
        if (mp==NULL) {
            GETB_PRT0(pInfo, "Mbuf allocation failed\n");
            return FALSE;
        }
#else
        MGETHDR(mp, M_DONTWAIT, MT_DATA);

        if (mp==NULL) {
            GETB_PRT0(pInfo, "Mbuf allocation failed\n");
            return FALSE;
        }

        MCLGET(mp, M_DONTWAIT);

        if ((mp->m_flags & M_EXT) == 0) {
             m_freem(mp);
             GETB_PRT0(pInfo, "Mbuf cluster allocation failed\n");
             return FALSE;
        }
#endif
    } else {
        mp = omp;
        mp->m_data = mp->m_ext.ext_buf;
    }

    mp->m_len = mp->m_pkthdr.len = MCLBYTES;

    pRDInfo->mbuf=mp;
    pRDInfo->mbuf_dma=(void*)vtophys(mtod(mp, vm_offset_t));

    pRD->rd0sts = 0;

    pRD->wBufLen=pInfo->rx_buf_sz;
    pRD->wBufLen |= RXDESC_I;
    pRD->dwBufAddrLo=(U32)pRDInfo->mbuf_dma;
    pRD->wBufAddrHi=0;

    return TRUE;
}

/**********************************************************************/
static void getb_setup_interface(PGETB_INFO pInfo)
{
    struct ifnet   *ifp;

    ifp = &pInfo->arpcom.ac_if;
#if __FreeBSD_version >= 502000
    if_initname(ifp, device_get_name(pInfo->dev), device_get_unit(pInfo->dev));
#else
    ifp->if_unit = pInfo->unit;
    ifp->if_name = "getb";
#endif
    ifp->if_mtu = ETHERMTU;
    ifp->if_output = ether_output;
    ifp->if_baudrate = 1000000000;
    ifp->if_init =  getb_init;
    ifp->if_softc = pInfo;
    ifp->if_flags = IFF_BROADCAST | IFF_SIMPLEX | IFF_MULTICAST;
    ifp->if_ioctl = getb_ioctl;
    ifp->if_start = getb_start;

#if __FreeBSD_version >= 400000
    if (pInfo->flags & (GETB_FLAGS_TX_CSUM|GETB_FLAGS_RX_CSUM)) {
        /* Set Checksum capability */
        ifp->if_hwassist = (CSUM_IP | CSUM_TCP | CSUM_UDP);
        ifp->if_capabilities |= IFCAP_HWCSUM;
    }

#if __FreeBSD_version >= 500000
    if (pInfo->flags & GETB_FLAGS_VLANTAG) {
        /* Set VLAN capability */
        ifp->if_capabilities |= IFCAP_VLAN_MTU|IFCAP_VLAN_HWTAGGING;
    }
#endif

    if (ifp->if_capabilities)
        ifp->if_capenable = ifp->if_capabilities;
#endif

    ifp->if_watchdog = getb_watchdog;
    ifp->if_snd.ifq_maxlen = (pInfo->sOpts.nTxDescs - 1)*7;

#if __FreeBSD_version >= 500000
    /*
     * Call MI attach routine.
     */
    ether_ifattach(ifp, pInfo->arpcom.ac_enaddr);

#elif __FreeBSD_version >= 410000 && __FreeBSD_version < 500000
    /*
     * Call MI attach routine.
     */
    ether_ifattach(ifp, ETHER_BPF_SUPPORTED);

#elif __FreeBSD_version >= 400000 && __FreeBSD_version < 410000
    /*
     * Call MI attach routines.
     */
    if_attach(ifp);
    ether_ifattach(ifp);
    bpfattach(ifp, DLT_EN10MB, sizeof(struct ether_header));

#else
    /*
     * Call MI attach routines.
     */
    if_attach(ifp);
    ether_ifattach(ifp);

#if NBPFILTER > 0
    bpfattach(ifp, DLT_EN10MB, sizeof(struct ether_header));
#endif

    at_shutdown(getb_shutdown, pInfo, SHUTDOWN_POST_SYNC);

#endif /*__FreeBSD_version*/


    ifmedia_init(&pInfo->media, IFM_IMASK, getb_media_change, getb_media_status);

    ifmedia_add(&pInfo->media, IFM_ETHER | IFM_10_T, 0, NULL);
    ifmedia_add(&pInfo->media, IFM_ETHER | IFM_10_T | IFM_FDX, 0, NULL);
    ifmedia_add(&pInfo->media, IFM_ETHER | IFM_100_TX, 0, NULL);
    ifmedia_add(&pInfo->media, IFM_ETHER | IFM_100_TX | IFM_FDX, 0, NULL);
    ifmedia_add(&pInfo->media, IFM_ETHER | IFM_1000_TX | IFM_FDX, 0, NULL);
    ifmedia_add(&pInfo->media, IFM_ETHER | IFM_1000_TX, 0, NULL);
    ifmedia_add(&pInfo->media, IFM_ETHER | IFM_AUTO, 0, NULL);
    ifmedia_set(&pInfo->media, IFM_ETHER | IFM_AUTO);

    return;
}


static void getb_media_status(struct ifnet *ifp, struct ifmediareq *ifmr)
{
    PGETB_INFO pInfo= ifp->if_softc;

    pInfo->mii_status=mac_check_media_mode(pInfo);

    ifmr->ifm_status = IFM_AVALID;
    ifmr->ifm_active = IFM_ETHER;

    if (pInfo->mii_status & GETB_LINK_FAIL)
        return;

    ifmr->ifm_status |= IFM_ACTIVE;

    if (pInfo->mii_status & GETB_SPEED_1000)
        ifmr->ifm_active |= IFM_1000_TX;
    else if (pInfo->mii_status & GETB_SPEED_100)
        ifmr->ifm_active |= IFM_100_TX;
    else
        ifmr->ifm_active |= IFM_10_T;

    if (pInfo->mii_status & GETB_DUPLEX_FULL)
        ifmr->ifm_active |= IFM_FDX;
    else
        ifmr->ifm_active |= IFM_HDX;
}


static int getb_media_change(struct ifnet *ifp)
{
    PGETB_INFO pInfo = ifp->if_softc;
    struct ifmedia  *ifm = &pInfo->media;
    int spd_dpx=0;

    if (IFM_TYPE(ifm->ifm_media) != IFM_ETHER)
        return(EINVAL);

    switch(IFM_SUBTYPE(ifm->ifm_media)) {
    case IFM_AUTO:
    case IFM_1000_TX:
        spd_dpx=SPD_DPX_AUTO;
        break;

    case IFM_100_TX:
        if ((ifm->ifm_media & IFM_GMASK) == IFM_FDX)
            spd_dpx=SPD_DPX_100_FULL;
        else
            spd_dpx=SPD_DPX_100_HALF;
        break;

    case IFM_10_T:
        if ((ifm->ifm_media & IFM_GMASK) == IFM_FDX)
            spd_dpx=SPD_DPX_10_FULL;
        else
            spd_dpx=SPD_DPX_10_HALF;
        break;

    default:
        GETB_PRT0(pInfo, "Unsupported media type\n");
    }

    pInfo->sOpts.spd_dpx=spd_dpx;
    getb_init(pInfo);

    return(0);
}


/*
  Initialiation of MAC registers
*/
static void getb_init_registers(PGETB_INFO pInfo, GETB_INIT_TYPE InitType)
{
    int         i, mii_status;

    mac_wol_reset(pInfo);

    switch (InitType) {
    case GETB_INIT_RESET:
    case GETB_INIT_WOL:

        mac_rx_queue_wake(pInfo);

        /*Fixed 10M HUB */
        BYTE_REG_BITS_ON(pInfo, CHIPGCR_LPSOPT, MAC_REG_CHIPGCR);

        mii_status=getb_get_opt_media_mode(pInfo);

        mii_init(pInfo, mii_status);

        if (getb_set_media_mode(pInfo, mii_status)!=GETB_LINK_CHANGE) {
            pInfo->mii_status=mac_check_media_mode(pInfo);
            getb_print_link_status(pInfo);
        }

        enable_flow_control_ability(pInfo);

        mac_clear_isr(pInfo);

        writel(pInfo, CR0_STOP, MAC_REG_CR0_CLR);
        writel(pInfo, (CR0_DPOLL|CR0_TXON|CR0_RXON|CR0_STRT), MAC_REG_CR0_SET);

        break;

    case GETB_INIT_COLD:
    default:
        /* do reset */
        getb_soft_reset(pInfo);

        for (i = 0; i < 6; i++)
            writeb(pInfo, pInfo->arpcom.ac_enaddr[i], MAC_REG_PAR+i);

        mac_set_rx_thresh(pInfo, pInfo->sOpts.rx_thresh);
        mac_set_dma_length(pInfo, pInfo->sOpts.DMA_length);

        /* suspend-well accept broadcast, multicast */
        writeb(pInfo, WOLCFG_SAM|WOLCFG_SAB, MAC_REG_WOLCFG_SET);

        /* back off algorithm use original IEEE standard */
        BYTE_REG_BITS_SET(pInfo, CFGB_OFSET,
            (CFGB_CRANDOM | CFGB_CAP | CFGB_MBA | CFGB_BAKOPT),
            MAC_REG_CFGB);

        /* Fixed 10M HUB */
        BYTE_REG_BITS_ON(pInfo, CHIPGCR_LPSOPT, MAC_REG_CHIPGCR);

        pInfo->IntMask=INT_MASK_DEF;

        /* load RD ring base address & size */
        writel(pInfo, (U32)pInfo->rd_pool_dma, MAC_REG_RDBASE_LO);
        writew(pInfo, pInfo->sOpts.nRxDescs-1, MAC_REG_RDCSIZE);

        /* run & wake RD ring */
        mac_rx_queue_run(pInfo);
        mac_rx_queue_wake(pInfo);

        /* load TD ring base address & size & run TD ring */
        writew(pInfo, pInfo->sOpts.nTxDescs-1, MAC_REG_TDCSIZE);
        for (i=0; i<pInfo->nTxQueues; i++) {
            writel(pInfo, (U32)pInfo->td_pool_dma[i], MAC_REG_TDBASE_LO+(i*4));
            mac_tx_queue_run(pInfo, i);
        }

        /* init cam filter */
        getb_init_cam_filter(pInfo);

        /* set packet filter
           receive directed and broadcast address */
        getb_set_multi(pInfo);

        enable_flow_control(pInfo);

        /* start mac */
        writel(pInfo, CR0_STOP, MAC_REG_CR0_CLR);
        writel(pInfo, (CR0_DPOLL|CR0_TXON|CR0_RXON|CR0_STRT), MAC_REG_CR0_SET);

        mii_status=getb_get_opt_media_mode(pInfo);

        /* init mii */
        mii_init(pInfo, mii_status);

        /* set media mode */
        getb_set_media_mode(pInfo, mii_status);

        enable_flow_control_ability(pInfo);

        mac_write_int_mask(pInfo, ISR_SRCI);
        mac_clear_isr(pInfo);
        mac_enable_int(pInfo);

    } /* switch */
}


static void getb_set_multi(PGETB_INFO pInfo)
{
    struct ifmultiaddr  *ifma;
    struct ifnet        *ifp = &pInfo->arpcom.ac_if;
    int                 mcnt = 0;
    U8                  rx_mode=(RCR_AM|RCR_AB);

    /* First, zot all the multicast entries. */
    bzero(pInfo->abyMCAMMask, sizeof(U8)*8);
    mac_set_cam_mask(pInfo, pInfo->abyMCAMMask, GETB_MULTICAST_CAM);
    writel(pInfo, 0, MAC_REG_MARCAM);
    writel(pInfo, 0, MAC_REG_MARCAM+4);

    if (ifp->if_flags & IFF_ALLMULTI || ifp->if_flags & IFF_PROMISC) {
        writel(pInfo, 0xffffffff, MAC_REG_MARCAM);
        writel(pInfo, 0xffffffff, MAC_REG_MARCAM+4);
        if (ifp->if_flags & IFF_PROMISC)
            rx_mode |= RCR_PROM;
    }
    else {

#if __FreeBSD_version < 500000
        LIST_FOREACH(ifma, &ifp->if_multiaddrs, ifma_link) {
#else
        TAILQ_FOREACH(ifma, &ifp->if_multiaddrs, ifma_link) {
#endif
            if (ifma->ifma_addr->sa_family != AF_LINK)
                continue;

            if (mcnt < pInfo->multicast_limit) {
                mac_set_cam(pInfo, mcnt,
                    LLADDR((struct sockaddr_dl *)ifma->ifma_addr),
                    GETB_MULTICAST_CAM);

                pInfo->abyMCAMMask[(mcnt)/8] |= 1<<((mcnt) & 7);
            } else {
                writel(pInfo, 0xffffffff, MAC_REG_MARCAM);
                writel(pInfo, 0xffffffff, MAC_REG_MARCAM+4);
                break;
            }
            mcnt++;
        }

        mac_set_cam_mask(pInfo, pInfo->abyMCAMMask, GETB_MULTICAST_CAM);
    }

    if (ifp->if_mtu > 1500)
        rx_mode |= RCR_AL;

    BYTE_REG_BITS_ON(pInfo, rx_mode, MAC_REG_RCR);

    return;
}


static BOOL getb_soft_reset(PGETB_INFO pInfo)
{
    int i=0;

    writel(pInfo, CR0_SFRST, MAC_REG_CR0_SET);

    for (i=0; i<W_MAX_TIMEOUT; i++) {
        udelay(5);
        if (!DWORD_REG_BITS_IS_ON(pInfo, CR0_SFRST, MAC_REG_CR0_SET))
            break;
    }

    if (i==W_MAX_TIMEOUT) {
        printf("getb%d: soft reset timed out!\n", pInfo->unit);
        writel(pInfo, CR0_FORSRST, MAC_REG_CR0_SET);
        /* delay 2ms*/
        mdelay(2);
    }

    mdelay(5);

    mac_eeprom_reload(pInfo);

    /* EEPROM reloaded will cause bit 0 in MAC_REG_CFGA turned on.
     * it makes MAC receive magic packet automatically. So, driver turn it off.
     */
    BYTE_REG_BITS_OFF(pInfo, CFGA_PACPI, MAC_REG_CFGA);

    return TRUE;
}


static void getb_init_cam_filter(PGETB_INFO pInfo)
{
    int  vid=0;

    /*Disable all cam*/
    bzero(pInfo->abyVCAMMask, sizeof(U8)*8);
    bzero(pInfo->abyMCAMMask, sizeof(U8)*8);
    mac_set_cam_mask(pInfo, pInfo->abyVCAMMask, GETB_VLAN_ID_CAM);
    mac_set_cam_mask(pInfo, pInfo->abyMCAMMask, GETB_MULTICAST_CAM);

    if (pInfo->flags & GETB_FLAGS_VLANTAG) {
        /* VLAN enabled */
        /* In FreeBSD, VLAN packets are filtered by SW, so disable HW filtering */
        /* set {PQEN, RTGOPT} = {1,0} */
        WORD_REG_BITS_SET(pInfo, (MCFG_PQEN), (MCFG_PQEN|MCFG_RTGOPT), MAC_REG_MCFG0);
    }
    else {
        /* Mode 1 */
        /* set {PQEN, RTGOPT} = {1,0} */
        WORD_REG_BITS_SET(pInfo, (MCFG_PQEN), (MCFG_PQEN|MCFG_RTGOPT), MAC_REG_MCFG0);
        
        /* Enable vcam 0 */
        mac_set_cam(pInfo, 0, (PU8)&vid, GETB_VLAN_ID_CAM);
        pInfo->abyVCAMMask[0]|=1;
        mac_set_cam_mask(pInfo, pInfo->abyVCAMMask, GETB_VLAN_ID_CAM);
        
        /* set VIDFR=1 to enable VLAN ID hardware filtering. */
        WORD_REG_BITS_ON(pInfo, MCFG_VIDFR, MAC_REG_MCFG0);    
    }
}


static U32 mac_check_media_mode(PGETB_INFO pInfo)
{
    U32                 status=0;
    U8                  byPHYSR0;
    U16                 wANAR;

    byPHYSR0=readb(pInfo, MAC_REG_PHYSR0);

    if (!(byPHYSR0 & PHYSR0_LINKGD))
        status|=GETB_LINK_FAIL;

    if (byPHYSR0 & PHYSR0_FDPX)
        status|=GETB_DUPLEX_FULL;

    if (byPHYSR0 & PHYSR0_SPDG)
        status|=GETB_SPEED_1000;

    if (byPHYSR0 & PHYSR0_SPD10)
        status|=GETB_SPEED_10;
    else
        status|=GETB_SPEED_100;

    if (MII_REG_BITS_IS_ON(pInfo, BMCR_AUTO, MII_REG_BMCR)) {
        getb_mii_read(pInfo, MII_REG_ANAR, &wANAR);
        if ((wANAR & (ANAR_TXFD|ANAR_TX|ANAR_10FD|ANAR_10))==
            (ANAR_TXFD|ANAR_TX|ANAR_10FD|ANAR_10)) {
            if (MII_REG_BITS_IS_ON(pInfo, G1000CR_1000|G1000CR_1000FD, MII_REG_G1000CR))
                status|=GETB_AUTONEG_ENABLE;
        }
    }

    return status;
}


/************************************************************************
* MII access , media link mode setting functions
************************************************************************/
static void mii_init(PGETB_INFO pInfo, U32 mii_status)
{
    U16 wBMCR;

    switch (PHYID_GET_PHY_ID(pInfo->dwPHYId)) {
    case PHYID_CICADA_CS8201:
        /*NOTICE!! For CICADA PHY
          Turn on AUX_MODE (bit 2) in MII AUX register (offset 1Ch) */
        MII_REG_BITS_ON(pInfo, AUXCR_MDPPS, MII_REG_AUXCR);

        /* Patch VT3119/VT3216 speed polling bug of MII Full-Duplex mode */
        if ((mii_status & (GETB_SPEED_100|GETB_SPEED_10)) && (mii_status & GETB_DUPLEX_FULL)) {
            MII_REG_BITS_ON(pInfo, TCSR_ECHODIS, MII_REG_TCSR);
        }
        else {
            MII_REG_BITS_OFF(pInfo, TCSR_ECHODIS, MII_REG_TCSR);
        }

        /* Turn on bit 2 in offset 0x1B, this patch code only adopt to CIS8201, not for integrated PHY in VT3216 */
        MII_REG_BITS_ON(pInfo, LED_LALBE, MII_REG_LED);

        break;

    case PHYID_VT3216_32BIT:
    case PHYID_VT3216_64BIT:

        /* Patch VT3119/VT3216 speed polling bug of MII Full-Duplex mode */
        if ((mii_status & (GETB_SPEED_100|GETB_SPEED_10)) && (mii_status & GETB_DUPLEX_FULL)) {
            MII_REG_BITS_ON(pInfo, TCSR_ECHODIS, MII_REG_TCSR);
        }
        else {
            MII_REG_BITS_OFF(pInfo, TCSR_ECHODIS, MII_REG_TCSR);
        }

        break;

    case PHYID_MARVELL_1000:
    case PHYID_MARVELL_1000S:
        MII_REG_BITS_ON(pInfo, 0x0B, 0x10);
        break;

    default:
        ;
    }

    getb_mii_read(pInfo, MII_REG_BMCR, &wBMCR);

    if (wBMCR & BMCR_ISO) {
        wBMCR &= ~BMCR_ISO;
        getb_mii_write(pInfo, MII_REG_BMCR, wBMCR);
    }
}


static void SafeDisableMiiAutoPoll (PGETB_INFO pInfo)
{
    WORD        ww;

    /* turn off MAUTO*/
    writeb(pInfo, 0, MAC_REG_MIICR);

    for (ww = 0; ww < W_MAX_TIMEOUT; ww++) {
        udelay(1);
        if (BYTE_REG_BITS_IS_ON(pInfo, MIISR_MIDLE, MAC_REG_MIISR))
            break;
    }
}


static void EnableMiiAutoPoll(PGETB_INFO pInfo)
{
    int ii;

    writeb(pInfo, 0, MAC_REG_MIICR);
    writeb(pInfo, MIIADR_SWMPL, MAC_REG_MIIADR);

    for (ii=0; ii<W_MAX_TIMEOUT; ii++) {
        udelay(1);
        if (BYTE_REG_BITS_IS_ON(pInfo, MIISR_MIDLE, MAC_REG_MIISR))
            break;
    }

    writeb(pInfo, MIICR_MAUTO, MAC_REG_MIICR);

    for (ii=0; ii<W_MAX_TIMEOUT; ii++) {
        udelay(1);
        if (!BYTE_REG_BITS_IS_ON(pInfo, MIISR_MIDLE, MAC_REG_MIISR))
            break;
    }

}


static BOOL getb_mii_read(PGETB_INFO pInfo, U8 byIdx, PU16 pdata)
{
    WORD        ww;

    /* disable MIICR_MAUTO, so that mii addr can be set normally*/
    SafeDisableMiiAutoPoll(pInfo);

    writeb(pInfo, byIdx, MAC_REG_MIIADR);

    BYTE_REG_BITS_ON(pInfo, MIICR_RCMD, MAC_REG_MIICR);

    for (ww = 0; ww < W_MAX_TIMEOUT; ww++) {
        if (!(readb(pInfo, MAC_REG_MIICR) & MIICR_RCMD))
            break;
    }

    *pdata=readw(pInfo, MAC_REG_MIIDATA);

    EnableMiiAutoPoll(pInfo);

    if (ww == W_MAX_TIMEOUT) {
        return FALSE;
    }
    return TRUE;
}


static BOOL getb_mii_write (PGETB_INFO pInfo, BYTE byMiiAddr, U16 wData)
{
    WORD        ww;

    /* disable MIICR_MAUTO, so that mii addr can be set normally*/
    SafeDisableMiiAutoPoll(pInfo);

    /* MII reg offset*/
    writeb(pInfo, byMiiAddr, MAC_REG_MIIADR);
    /* set MII data*/
    writew(pInfo, wData, MAC_REG_MIIDATA);

    /* turn on MIICR_WCMD*/
    BYTE_REG_BITS_ON(pInfo, MIICR_WCMD, MAC_REG_MIICR);

    /* W_MAX_TIMEOUT is the timeout period*/
    for (ww = 0; ww < W_MAX_TIMEOUT; ww++) {
        udelay(5);
        if (!(readb(pInfo, MAC_REG_MIICR) & MIICR_WCMD))
            break;
    }


    EnableMiiAutoPoll(pInfo);

    if (ww == W_MAX_TIMEOUT) {
        return FALSE;
    }

    return TRUE;
}


/*
  Get the media mode stored in EEPROM or module options
*/
static U32 getb_get_opt_media_mode(PGETB_INFO pInfo)
{
    U32         status=0;

    switch (pInfo->sOpts.spd_dpx) {
    case SPD_DPX_AUTO:
        status=GETB_AUTONEG_ENABLE;
        break;
    case SPD_DPX_100_FULL:
        status=GETB_SPEED_100|GETB_DUPLEX_FULL;
        break;
    case SPD_DPX_10_FULL:
        status=GETB_SPEED_10|GETB_DUPLEX_FULL;
        break;
    case SPD_DPX_100_HALF:
        status=GETB_SPEED_100;
        break;
    case SPD_DPX_10_HALF:
        status=GETB_SPEED_10;
        break;
    }

    return status;
}


static void mii_set_auto_on(PGETB_INFO pInfo)
{
    if (MII_REG_BITS_IS_ON(pInfo, BMCR_AUTO, MII_REG_BMCR))
        MII_REG_BITS_ON(pInfo, BMCR_REAUTO, MII_REG_BMCR);
    else
        MII_REG_BITS_ON(pInfo, BMCR_AUTO, MII_REG_BMCR);
}


static void mii_set_auto_off(PGETB_INFO pInfo)
{
    MII_REG_BITS_OFF(pInfo, BMCR_AUTO, MII_REG_BMCR);
}


static void set_mii_flow_control(PGETB_INFO pInfo)
{
    /*Enable or Disable PAUSE in ANAR*/
    switch(pInfo->sOpts.flow_cntl) {
    case FLOW_CNTL_DEFAULT:
        if (PHYID_GET_PHY_ID(pInfo->dwPHYId) == PHYID_CICADA_CS8201) {
            /* hardware default PAUSE/ASMDIR value in CIS8201 is (0,0) */
            MII_REG_BITS_OFF(pInfo, (ANAR_ASMDIR|ANAR_PAUSE), MII_REG_ANAR);

        }
        else if ((PHYID_GET_PHY_ID(pInfo->dwPHYId) == PHYID_VT3216_32BIT) ||
                 (PHYID_GET_PHY_ID(pInfo->dwPHYId) == PHYID_VT3216_64BIT)) {
            /* hardware default PAUSE/ASMDIR value in VT3216 is (1,1) */
            MII_REG_BITS_ON(pInfo, (ANAR_ASMDIR|ANAR_PAUSE), MII_REG_ANAR);
        }
        else {
            /* Do nothing */
        }
        break;

    case FLOW_CNTL_TX:
        MII_REG_BITS_OFF(pInfo, ANAR_PAUSE, MII_REG_ANAR);
        MII_REG_BITS_ON(pInfo, ANAR_ASMDIR, MII_REG_ANAR);
        break;

    case FLOW_CNTL_RX:
    case FLOW_CNTL_TX_RX:
        MII_REG_BITS_ON(pInfo, ANAR_PAUSE, MII_REG_ANAR);
        MII_REG_BITS_ON(pInfo, ANAR_ASMDIR, MII_REG_ANAR);
        break;

    case FLOW_CNTL_DISABLE:
        MII_REG_BITS_OFF(pInfo, ANAR_PAUSE, MII_REG_ANAR);
        MII_REG_BITS_OFF(pInfo, ANAR_ASMDIR, MII_REG_ANAR);
        break;
    default:
        break;
    }

}


static int getb_set_media_mode(PGETB_INFO pInfo, U32 mii_status)
{
    U32         curr_status;

    pInfo->mii_status=mii_check_media_mode(pInfo);
    curr_status= pInfo->mii_status & (~GETB_LINK_FAIL);

    /*Enable or Disable PAUSE in ANAR*/
    set_mii_flow_control(pInfo);

    /* clear force MAC mode bit */
    BYTE_REG_BITS_OFF(pInfo, CHIPGCR_FCMODE, MAC_REG_CHIPGCR);

    /* if connection type is AUTO    */
    if (mii_status & GETB_AUTONEG_ENABLE) {
        /* set duplex mode of MAC according to duplex mode of MII*/
        MII_REG_BITS_ON(pInfo, ANAR_TXFD|ANAR_TX|ANAR_10FD|ANAR_10, MII_REG_ANAR);
        MII_REG_BITS_ON(pInfo, G1000CR_1000FD|G1000CR_1000, MII_REG_G1000CR);
        MII_REG_BITS_ON(pInfo, BMCR_SPEED1G, MII_REG_BMCR);

        /* enable AUTO-NEGO mode*/
        mii_set_auto_on(pInfo);
    }
    else { /* Nway force */
        U16 wANAR;

        /* In forced mode, always turn-on MAC FCMODE */
        BYTE_REG_BITS_ON(pInfo, CHIPGCR_FCMODE, MAC_REG_CHIPGCR);

        mii_set_auto_off(pInfo);

        MII_REG_BITS_OFF(pInfo, G1000CR_1000FD|G1000CR_1000, MII_REG_G1000CR);
        MII_REG_BITS_OFF(pInfo, BMCR_SPEED1G, MII_REG_BMCR);

        getb_mii_read(pInfo, MII_REG_ANAR, &wANAR);

        wANAR &= (~(ANAR_TXFD|ANAR_TX|ANAR_10FD|ANAR_10));

        if (mii_status & GETB_SPEED_100) {
            if (mii_status & GETB_DUPLEX_FULL) {
                wANAR |=ANAR_TXFD;
                /*Turn on the forced mode of MAC*/
                BYTE_REG_BITS_ON(pInfo, CHIPGCR_FCFDX, MAC_REG_CHIPGCR);
            }
            else {
                wANAR |=ANAR_TX;
                /*Turn on the forced mode of MAC*/
                BYTE_REG_BITS_OFF(pInfo, CHIPGCR_FCFDX, MAC_REG_CHIPGCR);
            }
        }
        else {
            if (mii_status & GETB_DUPLEX_FULL) {
                wANAR |=ANAR_10FD;
                /*Turn on the forced mode of MAC*/
                BYTE_REG_BITS_ON(pInfo, CHIPGCR_FCFDX, MAC_REG_CHIPGCR);
            }
            else {
                wANAR |=ANAR_10;
                /*Turn on the forced mode of MAC*/
                BYTE_REG_BITS_OFF(pInfo, CHIPGCR_FCFDX, MAC_REG_CHIPGCR);
            }
        }

        getb_mii_write(pInfo, MII_REG_ANAR, wANAR);

        mii_set_auto_on(pInfo);

        getb_mii_read(pInfo, MII_REG_ANAR, &wANAR);
    }

    pInfo->mii_status=mii_check_media_mode(pInfo);

    return GETB_LINK_CHANGE;
}


static U32 mii_check_media_mode(PGETB_INFO pInfo)
{
    U32                 status=0;
    U16                 wANAR;

    if (!MII_REG_BITS_IS_ON(pInfo, BMSR_LNK, MII_REG_BMSR))
        status|=GETB_LINK_FAIL;

    if (MII_REG_BITS_IS_ON(pInfo, G1000CR_1000FD, MII_REG_G1000CR))
        status|=GETB_SPEED_1000|GETB_DUPLEX_FULL;
    else if (MII_REG_BITS_IS_ON(pInfo, G1000CR_1000, MII_REG_G1000CR))
        status|=(GETB_SPEED_1000);
    else {
        getb_mii_read(pInfo, MII_REG_ANAR, &wANAR);
        if (wANAR & ANAR_TXFD)
            status|=(GETB_SPEED_100|GETB_DUPLEX_FULL);
        else if (wANAR & ANAR_TX)
            status|=GETB_SPEED_100;
        else if (wANAR & ANAR_10FD)
            status|=(GETB_SPEED_10|GETB_DUPLEX_FULL);
        else
            status|=(GETB_SPEED_10);
    }

    if (MII_REG_BITS_IS_ON(pInfo, BMCR_AUTO, MII_REG_BMCR)) {
        getb_mii_read(pInfo, MII_REG_ANAR, &wANAR);
        if ((wANAR & (ANAR_TXFD|ANAR_TX|ANAR_10FD|ANAR_10))==
            (ANAR_TXFD|ANAR_TX|ANAR_10FD|ANAR_10)) {
            if (MII_REG_BITS_IS_ON(pInfo, G1000CR_1000|G1000CR_1000FD, MII_REG_G1000CR))
                status|=GETB_AUTONEG_ENABLE;
        }
    }

    return status;
}


static void getb_print_link_status(PGETB_INFO pInfo)
{
    U32     opt_media_mode=0;

    opt_media_mode=getb_get_opt_media_mode(pInfo);

    if (pInfo->mii_status & GETB_LINK_FAIL) {
        GETB_PRT0(pInfo, "Failed to detect cable link\n");
    }
    else {
        if (pInfo->mii_status & GETB_AUTONEG_ENABLE)
            GETB_PRT0(pInfo, "Link autonegation");
        else
            GETB_PRT0(pInfo, "Link forced");

        if (opt_media_mode & GETB_AUTONEG_ENABLE) {
            if (pInfo->mii_status & GETB_SPEED_1000)
                printf(" speed 1000M bps");
            else if (pInfo->mii_status & GETB_SPEED_100)
                printf(" speed 100M bps");
            else
                printf(" speed 10M bps");
        }
        /*
          Set variable to correct speed mode
          Patch VT3119/VT3216 speed polling bug of MII Full-Duplex mode
         */
        else {
            if (opt_media_mode & GETB_SPEED_100)
                printf(" speed 100M bps");
            else
                printf(" speed 10M bps");
        }

        if (pInfo->mii_status & GETB_DUPLEX_FULL)
            printf(" full duplex\n");
        else
            printf(" half duplex\n");

    }
}


static void enable_flow_control_ability(PGETB_INFO pInfo)
{

    switch(pInfo->sOpts.flow_cntl) {
    case FLOW_CNTL_DEFAULT:
        /* enable/disable RX flow control according to the result of negotiation */
        if (BYTE_REG_BITS_IS_ON(pInfo, PHYSR0_RXFLC, MAC_REG_PHYSR0))
            writel(pInfo, CR0_FDXRFCEN, MAC_REG_CR0_SET);
        else
            writel(pInfo, CR0_FDXRFCEN, MAC_REG_CR0_CLR);

        /* enable/disable TX flow control according to the result of negotiation */
        if (BYTE_REG_BITS_IS_ON(pInfo, PHYSR0_TXFLC, MAC_REG_PHYSR0))
            writel(pInfo, CR0_FDXTFCEN, MAC_REG_CR0_SET);
        else
            writel(pInfo, CR0_FDXTFCEN, MAC_REG_CR0_CLR);

        break;

    case FLOW_CNTL_TX:
        /* force MAC send PAUSE frame when RD is not enough */
        writel(pInfo, CR0_FDXRFCEN, MAC_REG_CR0_CLR);
        writel(pInfo, CR0_FDXTFCEN, MAC_REG_CR0_SET);
        break;

    case FLOW_CNTL_RX:
        /* force MAC response to PAUSE frame from link partner */
        writel(pInfo, CR0_FDXRFCEN, MAC_REG_CR0_SET);
        writel(pInfo, CR0_FDXTFCEN, MAC_REG_CR0_CLR);
        break;

    case FLOW_CNTL_TX_RX:
        writel(pInfo, (CR0_FDXRFCEN | CR0_FDXTFCEN), MAC_REG_CR0_SET);
        break;

    case FLOW_CNTL_DISABLE:
        writel(pInfo, (CR0_FDXRFCEN | CR0_FDXTFCEN), MAC_REG_CR0_CLR);
        break;

    }

}



